;***************************************************************************
;  YAPP21.ASM
;
;  YAPP! - Yet Another Pic Programmer
;
;  PIC serial in-circuit programmer
;***************************************************************************
;
;  (c) Sergio Tanzilli - E-mail: tanzilli@picpoint.com
;                        Web   : http://www.picpoint.com
;
;  Assembled with Microchip(tm) MPASM 2.01
;
;  Releases story:
;
;  2.0  July 16,     1998    PIC16C54/56/16F84 host processor source version
;  2.01 July 17,     1998    Fixed transient problem on clock line
;  2.1  September 2, 1998    Added SET_PROG_TIME command
;                            Added END_PROGR command
;***************************************************************************

;***************************************************************************
; This source has been tested on PIC16C54, 16C56, 16C84 and 16F84 as host
; processor on YAPP! programmer
;
; Please program the YAPP! host processor with the following configuration:
;
; Osc=XT
; Watch dog timer = OFF
; Power up timer = ON (only on 16F84, 16C84)
;***************************************************************************


;Select processor: 16C54, 16C56 or 16F84 (valid also for PIC16C84)

                PROCESSOR       16C54
                RADIX           DEC

                IFDEF __16C54
                INCLUDE         "P16C5X.INC"
                ENDIF

                IFDEF __16C56
                INCLUDE         "P16C5X.INC"
                ENDIF

                IFDEF __16F84
                INCLUDE         "P16F84.INC"
                ENDIF

                ;Use of I/O line

                ;RA0 (pin 17) --> Ready green led (1 = On)
                ;RA1 (pin 18) --> Programming red led (1 = On)
                ;RA2 (pin  1) <-- Not used
                ;RA3 (pin  2) <-- Not used

                ;RB0 (pin  6) --> VPP enable
                ;RB1 (pin  7) --> VCC enable
                ;RB2 (pin  8) --> DATA OUT
                ;RB3 (pin  9) <-- DATA IN
                ;RB4 (pin 10) --> CLOCK OUT
                ;RB5 (pin 11) --> TxD
                ;RB6 (pin 12) <-- RxD
                ;RB7 (pin 13) <-- Not used


;Definition of I/O line on port A

READY_LED       equ     0
PROG_LED        equ     1

;Definition of I/O line on port B

SEND_VPP        equ     0               ;Enable VPP on MCLR
SEND_GND        equ     1               ;Enable GND on MCLR
PIC_DATA_OUT    equ     2               ;Target data out (RB7 on PIC target)
PIC_DATA_IN     equ     3               ;Target data in (RB7 on PIC target)
PIC_CLOCK       equ     4               ;Target clock (RB6 on PIC target)
TX              equ     5               ;Tx data
RX              equ     6               ;Rx data

;Acknowledge commands (both YAPP! and Target PIC in TMOD1 mode)

LOAD_CONFIG     equ     0x00
LOAD_DATA       equ     0x02
LOAD_DATA_DM    equ     0x03
READ_DATA       equ     0x04
READ_DATA_DM    equ     0x05
INC_ADDR        equ     0x06
BEGIN_PROG      equ     0x08
END_PROG        equ     0x0E

;Acknowledge commands (only YAPP!)

RESET_MODE      equ     64
TMOD1_MODE      equ     65
RUN_MODE        equ     66

;Other acknowledge commands (only YAPP!, non standard Microchip)

SET_CLOCK_HI    equ     67
SET_CLOCK_LO    equ     68
SET_DATA_HI     equ     69
SET_DATA_LO     equ     70
SET_PROG_TIME   equ     71      ;Set programming delay time

;**********************************************************************
; Clock frequency related constant (4 MHz)
;**********************************************************************

BIT_DELAY       equ    23               ;Baud rate: 9600 baud

;**********************************************************************
; Program register
;**********************************************************************

                IFDEF __16C54
                org     0x08
                ENDIF

                IFDEF __16C56
                org     0x08
                ENDIF

                IFDEF __16F84
                org     0x0C
                ENDIF

;Register used for RS232 rx/tx

ShiftReg        res     1
BitCount        res     1
ByteCount       res     1

TmpRegister     res     1
PicCmd          res     1

PicDataLo       res     1
PicDataHi       res     1

ProgTime        res     1
ProgTimeCounter res     1

;**********************************************************************
; MACRO - Delay subroutine with watch dog timer clearing
;
; Macro parameters:
;
;   VALUE:      Delay obtained = ((VALUE-1)*4+5)*(1/(Fosc/4))
;
;**********************************************************************

DELAY           MACRO   VALUE
                LOCAL   REDO

                movlw   VALUE
                movwf   TmpRegister
REDO

                clrwdt                  ;Clear watch dog timer

                decfsz  TmpRegister,F
                goto    REDO

                ENDM

DATA_OUT_HI     MACRO
                bsf     PORTB,PIC_DATA_OUT
                ENDM

DATA_OUT_LO     MACRO
                bcf     PORTB,PIC_DATA_OUT
                ENDM

CLOCK_HI        MACRO
                bsf     PORTB,PIC_CLOCK
                ENDM

CLOCK_LO        MACRO
                bcf     PORTB,PIC_CLOCK
                ENDM

READY_LED_ON    MACRO
                bsf     PORTA,READY_LED
                bcf     PORTA,PROG_LED
                ENDM

PROG_LED_ON     MACRO
                bcf     PORTA,READY_LED
                bsf     PORTA,PROG_LED
                ENDM


;**************************************************************************
; Reset vector for 16F84
;**************************************************************************

                IFDEF __16F84
                org     0x00
                goto    Main
                ENDIF

;**************************************************************************
; Send a character on RS232
; (9600 baud,8 data bit, 1 stop bit, No parity)
;
; Input: W = Character to send
;**************************************************************************

TxChar
                movwf   ShiftReg

                movlw   8               ;Data lenght
                movwf   BitCount

                bcf     PORTB,TX        ;Send start bit
                nop
                nop
                nop
                nop
                nop
                nop
                nop

                DELAY   BIT_DELAY

                ;Tx loop

TxLoop
                btfss   ShiftReg,0
                goto    TxLo

                nop
                bsf     PORTB,TX
                goto    cTx
TxLo
                bcf     PORTB,TX
                goto    cTx
cTx
                nop
                rrf     ShiftReg,F

                DELAY   BIT_DELAY

                decfsz  BitCount,F
                goto    TxLoop
                nop
                nop
                nop
                nop

                bsf     PORTB,TX                ;Stop bit
                DELAY   BIT_DELAY
                DELAY   2
                nop

                bsf     PORTB,TX
                DELAY   BIT_DELAY
                DELAY   2
                return


;**************************************************************************
; Receive a character from RS232
; (9600 baud,8 data bit, 1 stop bit, No parity)
;
; Return code:
;
;   ShiftReg:  Received character
;   W:         0 = Ok
;              1 = Timeout error
;              2 = Frame error (stop bit = 0)
;**************************************************************************

RxChar
                clrf    BitCount
WaitStartBit
                btfss   PORTB,RX                ;Wait start bit until timeout
                goto    ReceivedStartBit

                movlw   250
                xorwf   BitCount,W
                btfsc   STATUS,Z
                retlw   1                       ;Timeout error

                incf    BitCount,F
                goto    WaitStartBit

ReceivedStartBit

                clrf    ShiftReg

                movlw   8                       ;Data lenght
                movwf   BitCount

                DELAY   BIT_DELAY+BIT_DELAY/5

                ;Loop di lettura dei bit dati
wDB
                btfss   PORTB,RX
                goto    RxBitL
RxBitH
                nop
                bsf     STATUS,C
                goto    RxShift
RxBitL
                bcf     STATUS,C
                goto    RxShift
RxShift
                nop
                rrf     ShiftReg,F

                DELAY   BIT_DELAY

                decfsz  BitCount,F
                goto    wDB
                nop
                nop
                nop
                nop

                btfss   PORTB,RX
                retlw   2               ;Frame error
                retlw   0               ;Ok

;**********************************************************************
; Put down MCLR pin on target PIC
;**********************************************************************

MCLR_GND
                bcf     PORTB,SEND_VPP
                bsf     PORTB,SEND_GND
                retlw   0

;**********************************************************************
; Send VPP on MCLR pin of target PIC
;**********************************************************************

MCLR_VPP
                bsf     PORTB,SEND_VPP
                bcf     PORTB,SEND_GND
                retlw   0

;**********************************************************************
; Send VCC on MCLR pin of target PIC
;**********************************************************************

MCLR_VCC
                bcf     PORTB,SEND_VPP
                bcf     PORTB,SEND_GND
                retlw   0

;**********************************************************************
; Send a ACK character to PC
;**********************************************************************

SendAck
                movlw   '0'
                call    TxChar
                retlw   0

;**********************************************************************
; Send a command to target PIC
;
; PicCmd = Comando da trasmettere
;**********************************************************************

SendCmdToPic
                movlw   6
                movwf   BitCount
SendLoop
                btfss   PicCmd,0
                goto    DataLo

                DATA_OUT_HI
                goto    NextDataBit
DataLo
                DATA_OUT_LO
NextDataBit
                CLOCK_HI
                CLOCK_LO

                rrf     PicCmd,F
                decfsz  BitCount,F
                goto    SendLoop

                retlw   0

;**********************************************************************
; Receive a word (14 bit) form target PIC
;
; Return code:
;
; PicDataLo, PicDataHi
;**********************************************************************

PicReadData
                DATA_OUT_HI

                CLOCK_HI
                CLOCK_LO

                ;Read lower bit

                movlw   8
                movwf   BitCount
LoReadLoop
                CLOCK_HI
                CLOCK_LO

                bsf     STATUS,C
                btfss   PORTB,PIC_DATA_IN
                bcf     STATUS,C

                rrf     PicDataLo,F

                decfsz  BitCount,F
                goto    LoReadLoop

                ;Read high bit

                movlw   6
                movwf   BitCount
HiReadLoop
                CLOCK_HI
                CLOCK_LO

                bsf     PicDataHi,6
                btfss   PORTB,PIC_DATA_IN
                bcf     PicDataHi,6

                rrf     PicDataHi,F

                decfsz  BitCount,F
                goto    HiReadLoop

                CLOCK_HI
                CLOCK_LO

                bcf     PicDataHi,6
                bcf     PicDataHi,7

                return

;**********************************************************************
; Send a word (14 bit) on target PIC
;
; Parameters:
;
; PicDataLo, PicDataHi
;**********************************************************************

PicWriteData
                CLOCK_HI
                CLOCK_LO

                ;Send lower bits

                movlw   8
                movwf   BitCount
LoWriteLoop
                CLOCK_HI

                btfss   PicDataLo,0
                goto    SendLo_1
                goto    SendHi_1
SendLo_1
                DATA_OUT_LO
                goto    _Send_1
SendHi_1
                DATA_OUT_HI
                goto    _Send_1
_Send_1
                CLOCK_LO

                rrf     PicDataLo,1

                decfsz  BitCount,1
                goto    LoWriteLoop

                ;Send highter bits

                movlw   6
                movwf   BitCount
HiWriteLoop
                CLOCK_HI

                btfss   PicDataHi,0
                goto    SendLo_2
                goto    SendHi_2
SendLo_2
                DATA_OUT_LO
                goto    _Send_2
SendHi_2
                DATA_OUT_HI
                goto    _Send_2
_Send_2
                CLOCK_LO

                rrf     PicDataHi,1

                decfsz  BitCount,1
                goto    HiWriteLoop

                CLOCK_HI
                CLOCK_LO

                return


;**********************************************************************
; Main
;**********************************************************************

Main
                IFDEF   __16C54
                movlw   00011100B       ;Line definition port A
                tris    PORTA

                movlw   01001000B       ;Line definition port B
                tris    PORTB

                movlw   00000100B
                option
                ENDIF

                IFDEF   __16C56
                movlw   00011100B       ;Line definition port A
                tris    PORTA

                movlw   01001000B       ;Line definition port B
                tris    PORTB

                movlw   00000100B
                option
                ENDIF

                IFDEF   __16F84
                bsf     STATUS,RP0      ;Swap to bank 1

                movlw   00011100B       ;Line definition port A
                movwf   TRISA

                movlw   01001000B       ;Line definition port B
                movwf   TRISB

                movlw   10000100B
                movwf   OPTION_REG

                movlw   00000000B       ;Disable interrupt
                movwf   INTCON

                bcf     STATUS,RP0      ;Swap to bank 1
                ENDIF

                ;Initial status of line

                bsf     PORTB,TX

                movlw   100             ;Set program time to 10ms (FLASH)
                movwf   ProgTime

                READY_LED_ON            ;Ready led on

                ;Set target PIC in run mode

                DATA_OUT_HI
                CLOCK_HI
                call    MCLR_VCC

;**********************************************************************
; Main loop
;**********************************************************************


MainLoop
                btfsc   PORTB,RX        ;Received start bit from PC ?
                goto    MainLoop        ;No, wait.

;**********************************************************************
; Receive command from PC
;**********************************************************************

StartBit
                call    RxChar

                xorlw   0
                btfss   STATUS,Z        ;Timeout or frame error ?
                goto    MainLoop        ;Yes, abort...

                ;Check command received from PC

                ;----------------------------------
                ; Program mode MCLR = VPP
                ;----------------------------------

CheckProgramMode
                movlw   TMOD1_MODE
                xorwf   ShiftReg,W
                bnz     _CheckProgramMode

                PROG_LED_ON
                DATA_OUT_LO
                CLOCK_LO
                call    MCLR_VPP

                call    SendAck
                goto    MainLoop

_CheckProgramMode

                ;----------------------------------
                ; Reset mode MCLR = 0
                ;----------------------------------

CheckResetMode
                movlw   RESET_MODE
                xorwf   ShiftReg,W
                bnz     _CheckResetMode

                READY_LED_ON
                DATA_OUT_LO
                CLOCK_LO
                call    MCLR_GND

                call    SendAck
                goto    MainLoop

_CheckResetMode

                ;----------------------------------
                ; Run mode MCLR = VCC
                ;----------------------------------

CheckRunMode
                movlw   RUN_MODE
                xorwf   ShiftReg,W
                bnz     _CheckRunMode

                READY_LED_ON
                DATA_OUT_HI
                CLOCK_HI
                call    MCLR_VCC

                call    SendAck
                goto    MainLoop

_CheckRunMode


                ;----------------------------------
                ; Set clock line hi
                ;----------------------------------

CheckSetClockHi
                movlw   SET_CLOCK_HI
                xorwf   ShiftReg,W
                bnz     _CheckSetClockHi

                CLOCK_HI

                call    SendAck
                goto    MainLoop

_CheckSetClockHi

                ;----------------------------------
                ; Set clock line low
                ;----------------------------------

CheckSetClockLo
                movlw   SET_CLOCK_LO
                xorwf   ShiftReg,W
                bnz     _CheckSetClockLo

                CLOCK_LO

                call    SendAck
                goto    MainLoop

_CheckSetClockLo

                ;----------------------------------
                ; Set data line hi
                ;----------------------------------

CheckSetDataHi
                movlw   SET_DATA_HI
                xorwf   ShiftReg,W
                bnz     _CheckSetDataHi

                DATA_OUT_HI

                call    SendAck
                goto    MainLoop

_CheckSetDataHi

                ;----------------------------------
                ; Set data line low
                ;----------------------------------

CheckSetDataLo
                movlw   SET_DATA_LO
                xorwf   ShiftReg,W
                bnz     _CheckSetDataLo

                DATA_OUT_LO

                call    SendAck
                goto    MainLoop

_CheckSetDataLo

                ;----------------------------------
                ; Set programming delay time
                ;----------------------------------

CheckSetProgTime
                movlw   SET_PROG_TIME
                xorwf   ShiftReg,W
                bnz     _CheckSetProgTime

                call    RxChar          ;Get programming time (1 = 100us)

                xorlw   0
                btfss   STATUS,Z        ;Timeout or frame error ?
                goto    MainLoop        ;Yes, abort...

                movf    ShiftReg,W
                movwf   ProgTime

                call    SendAck
                goto    MainLoop

_CheckSetProgTime


;**********************************************************************
; Every other command received will be send to target pic
;**********************************************************************

                movf    ShiftReg,W
                movwf   PicCmd
                call    SendCmdToPic

                ;-----------------------------------------------------
                ; For LOAD_DATA, LOAD_CONFIG and LOAD_DATA_DM
                ; read a word from PC and send it to target pic
                ;-----------------------------------------------------

CheckLoadData
                movlw   LOAD_DATA
                xorwf   ShiftReg,W
                bz      __CheckLoadData

                movlw   LOAD_DATA_DM
                xorwf   ShiftReg,W
                bz      __CheckLoadData

                movlw   LOAD_CONFIG
                xorwf   ShiftReg,W
                bnz     _CheckLoadData

__CheckLoadData
                call    RxChar          ;Lower byte

                xorlw   0
                btfss   STATUS,Z        ;Timeout or frame error ?
                goto    MainLoop        ;Yes, abort...

                movf    ShiftReg,W      ;No, store byte received
                movwf   PicDataLo

                call    RxChar          ;Highter byte

                xorlw   0
                btfss   STATUS,Z        ;Timeout or frame error ?
                goto    MainLoop        ;Yes abort...

                movf    ShiftReg,W      ;No, store byte received
                movwf   PicDataHi

                call    PicWriteData    ;Send value to pic

                call    SendAck

                goto    MainLoop

_CheckLoadData

                ;-----------------------------------------------------
                ; INC_ADDR
                ;-----------------------------------------------------

CheckIncAddr
                movlw   INC_ADDR
                xorwf   ShiftReg,W
                bnz     _CheckIncAddr

                call    SendAck
                goto    MainLoop

_CheckIncAddr

                ;-----------------------------------------------------
                ; BEGIN_PROG (wait ProgTime*100uS before send ACK)
                ;-----------------------------------------------------

CheckBeginProg
                movlw   BEGIN_PROG
                xorwf   ShiftReg,W
                bnz     _CheckBeginProg

                movf    ProgTime,W
                movwf   ProgTimeCounter

ProgrammingLoop
                DELAY   25

                decfsz  ProgTimeCounter,F
                goto    ProgrammingLoop

                call    SendAck
                goto    MainLoop

_CheckBeginProg

                ;-----------------------------------------------------
                ; READ_DATA
                ; Read word from PIC and send it to PC
                ;-----------------------------------------------------

CheckReadData
                movlw   READ_DATA
                xorwf   ShiftReg,W
                bnz     _CheckReadData

                call    PicReadData

                movf    PicDataLo,W
                call    TxChar

                movf    PicDataHi,W
                call    TxChar

                goto    MainLoop

_CheckReadData

                ;For every other commands send an ACK to PC

                call    SendAck
                goto    MainLoop

;**************************************************************************
; Reset vector for 16C54
;**************************************************************************

                IFDEF __16C54
                org     0x1FF
                goto    Main
                ENDIF

;**************************************************************************
; Reset vector for 16C56
;**************************************************************************

                IFDEF __16C56
                org     0x3FF
                goto    Main
                ENDIF

                end

